from anyio import create_task_group, sleep
from time import monotonic


class Monitor:
    def __init__(self, period: float = 0.01):
        self._period = period
        self._result = 0
        self._iter = 1

    async def __aenter__(self) -> "Monitor":
        self._task_group = create_task_group()
        tg = await self._task_group.__aenter__()
        tg.start_soon(self._run)
        self._cancel_scope = tg.cancel_scope
        return self

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        self._cancel_scope.cancel()
        await self._task_group.__aexit__(None, None, None)

    async def _run(self):
        while True:
            t0 = monotonic()
            await sleep(self._period)
            t1 = monotonic()
            factor = (t1 - t0) / self._period
            self._result = self._result + (factor - self._result) / self._iter
            self._iter += 1

    @property
    def result(self) -> float:
        return self._result



import pytest
from anyioutils import ALL_COMPLETED, FIRST_COMPLETED, FIRST_EXCEPTION, Task, create_task, wait
from anyio import create_task_group, sleep

pytestmark = pytest.mark.anyio


async def foo(wait: float = 0, raise_exception: bool = False) -> str:
    await sleep(wait)
    if raise_exception:
        raise RuntimeError("foo")
    return "foo"


async def test_wait_all_completed() -> None:
    async with create_task_group() as tg:
        tasks = [Task(aw) for aw in (foo(), foo())]
        done, pending = await wait(tasks, tg, return_when=ALL_COMPLETED)
        assert done == set(tasks)
        assert not pending
        for task in done:
            assert task.result() == "foo"


async def test_wait_first_completed() -> None:
    async with create_task_group() as tg:
        tasks = [Task(aw) for aw in (foo(), foo(wait=0.1))]
        done, pending = await wait(tasks, tg, return_when=FIRST_COMPLETED)
        assert len(done) == 1
        assert len(pending) == 1
        assert done == set([tasks[0]]) 
        assert pending == set([tasks[1]]) 
        for task in done:
            assert task.result() == "foo"
        for task in pending:
            assert await task.wait() == "foo"


async def test_wait_first_exception() -> None:
    async with create_task_group() as tg:
        tasks = [Task(aw) for aw in (foo(raise_exception=True), foo(wait=0.1))]
        done, pending = await wait(tasks, tg, return_when=FIRST_EXCEPTION)
        assert len(done) == 1
        assert len(pending) == 1
        assert done == set([tasks[0]]) 
        assert pending == set([tasks[1]]) 
        for task in done:
            with pytest.raises(RuntimeError) as excinfo:
                task.result()
            assert str(excinfo.value) == "foo"
        for task in pending:
            assert await task.wait() == "foo"


async def test_wait_not_tasks() -> None:
    async with create_task_group() as tg:
        aw = foo()
        with pytest.raises(TypeError) as excinfo:
            done, pending = await wait([aw], tg)
        assert str(excinfo.value) == "Pass tasks, not <class 'coroutine'>"
        await aw



from __future__ import annotations

from contextlib import AsyncExitStack
from typing import Any, Coroutine, TypeVar

from anyio import CancelScope, create_task_group

from ._task import Task, _task_group, create_task as _create_task

T = TypeVar("T")


class TaskGroup:
    @property
    def cancel_scope(self) -> CancelScope:
        return self._task_group.cancel_scope

    async def __aenter__(self) -> "TaskGroup":
        async with AsyncExitStack() as exit_stack:
            self._task_group = await exit_stack.enter_async_context(create_task_group())
            self._token = _task_group.set(self._task_group)
            self._exit_stack = exit_stack.pop_all()
        return self

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        _task_group.reset(self._token)
        return await self._exit_stack.__aexit__(exc_type, exc_value, exc_tb)

    def create_task(self, coro: Coroutine[Any, Any, T], *, name: str | None = None) -> Task[T]:
        return _create_task(coro, self._task_group, name=name)



"""Utility classes and functions for AnyIO."""

from ._event import Event as Event
from ._exceptions import CancelledError as CancelledError
from ._exceptions import InvalidStateError as InvalidStateError
from ._future import Future as Future
from ._guest import start_guest_run as start_guest_run
from ._monitor import Monitor as Monitor
from ._queue import Queue as Queue
from ._task import Task as Task
from ._task import create_task as create_task
from ._task_group import TaskGroup as TaskGroup
from ._wait import ALL_COMPLETED as ALL_COMPLETED
from ._wait import FIRST_COMPLETED as FIRST_COMPLETED
from ._wait import FIRST_EXCEPTION as FIRST_EXCEPTION
from ._wait import wait as wait



from sys import version_info

import pytest
from anyioutils import CancelledError, InvalidStateError, Task, create_task
from anyio import Event, create_task_group, sleep

if version_info < (3, 11):
    from exceptiongroup import BaseExceptionGroup, ExceptionGroup  # pragma: no cover

pytestmark = pytest.mark.anyio


async def test_task_result1():
    event = Event()

    async def foo():
        event.set()
        return 1

    async with create_task_group() as tg:
        task = Task(foo())
        with pytest.raises(InvalidStateError):
            task.result()
        tg.start_soon(task.wait)
        await event.wait()
        assert await task.wait() == 1
        assert task.result() == 1


async def test_task_result2():
    async def foo():
        return 1

    async with create_task_group() as tg:
        task = create_task(foo(), tg)
        assert await task.wait() == 1


async def test_exception():
    async def foo():
        raise RuntimeError

    task = Task(foo())
    with pytest.raises(InvalidStateError):
        task.exception()

    async with create_task_group() as tg:
        for _ in range(3):
            with pytest.raises(RuntimeError):
                await task.wait()
            assert task.done()
            assert type(task.exception()) == RuntimeError
            with pytest.raises(RuntimeError):
                task.result()


async def test_task_cancelled1():
    event = Event()

    async def bar():
        event.set()
        await sleep(float("inf"))

    with pytest.raises(BaseExceptionGroup) as excinfo:
        async with create_task_group() as tg:
            task = create_task(bar(), tg)
            await event.wait()
            task.cancel(raise_exception=True)
            assert task.cancelled()
            with pytest.raises(CancelledError):
                task.exception()
    assert excinfo.group_contains(CancelledError)

    with pytest.raises(CancelledError):
        task.result()


async def test_task_cancelled2():
    event = Event()

    async def bar():
        event.set()
        await sleep(float("inf"))

    with pytest.raises(BaseExceptionGroup) as excinfo:
        async with create_task_group() as tg:
            task = create_task(bar(), tg)
            await event.wait()
            task.cancel(raise_exception=True)
            await task.wait()
    assert excinfo.group_contains(CancelledError)


async def test_task_cancelled3():
    event = Event()

    async def bar():
        event.set()
        await sleep(float("inf"))

    async with create_task_group() as tg:
        task = create_task(bar(), tg)
        await event.wait()
        task.cancel()
        assert await task.wait() is None


async def test_callback():
    async def foo():
        pass

    task0 = Task(foo())
    callback0_called = False

    def callback0(task):
        nonlocal callback0_called
        assert task == task0
        callback0_called = True

    task0.add_done_callback(callback0)
    await task0.wait()
    assert callback0_called

    task1 = Task(foo())
    callback1_called = False

    def callback1(task):
        nonlocal callback1_called
        assert task == task1  # pragma: no cover
        callback1_called = True  # pragma: no cover

    task1.add_done_callback(callback1)
    task1.remove_done_callback(callback1)
    await task1.wait()
    assert not callback1_called

    task2 = Task(foo())

    def callback2(f):
        raise RuntimeError()

    task2.add_done_callback(callback2)

    with pytest.raises(ExceptionGroup):
        await task2.wait()

    task3 = Task(foo())

    def callback3(f):
        raise RuntimeError()

    task3.add_done_callback(callback3)
    task3.add_done_callback(callback3)

    with pytest.raises(ExceptionGroup):
        await task3.wait()


async def test_add_done_callback_already_done():
    async def foo():
        pass

    task = Task(foo())
    await task.wait()
    callback_called = False

    def callback(future):
        nonlocal callback_called
        callback_called = True
        raise RuntimeError()

    with pytest.raises(RuntimeError):
        task.add_done_callback(callback)

    assert callback_called



from sys import version_info

import pytest
from anyioutils import CancelledError, Future, InvalidStateError
from anyio import create_task_group

if version_info < (3, 11):
    from exceptiongroup import ExceptionGroup  # pragma: no cover

pytestmark = pytest.mark.anyio


async def test_result():
    future = Future()
    with pytest.raises(InvalidStateError):
        future.result()

    async def set_result():
        future.set_result(1)

    async with create_task_group() as tg:
        tg.start_soon(set_result)
        for _ in range(3):
            assert await future.wait() == 1
            assert future.done()
            assert future.result() == 1
        with pytest.raises(InvalidStateError):
            future.set_result(2)


async def test_exception():
    future = Future()
    with pytest.raises(InvalidStateError):
        future.exception()

    async def set_exception():
        future.set_exception(RuntimeError)

    async with create_task_group() as tg:
        tg.start_soon(set_exception)
        for _ in range(3):
            with pytest.raises(RuntimeError):
                await future.wait()
            assert future.done()
            assert future.exception() is RuntimeError
            with pytest.raises(RuntimeError):
                future.result()
        with pytest.raises(InvalidStateError):
            future.set_exception(RuntimeError)


async def test_cancel():
    future = Future()
    assert not future.cancelled()

    async def cancel():
        assert future.cancel(raise_exception=True)

    async with create_task_group() as tg:
        tg.start_soon(cancel)
        for _ in range(3):
            with pytest.raises(CancelledError):
                await future.wait()
            assert future.done()
            assert future.cancelled()
            with pytest.raises(CancelledError):
                assert future.exception()
            with pytest.raises(CancelledError):
                future.result()


async def test_immediate_cancel():
    future = Future()
    assert not future.cancelled()
    assert future.cancel(raise_exception=True)
    assert future.cancelled()
    assert future.done()
    for _ in range(3):
        with pytest.raises(CancelledError):
            await future.wait()
        with pytest.raises(CancelledError):
            assert future.exception()
        with pytest.raises(CancelledError):
            future.result()


def test_cancel_already_done():
    future = Future()
    future.set_result(0)
    assert not future.cancel()


async def test_callback():
    future0 = Future()
    callback0_called = False

    def callback0(future):
        nonlocal callback0_called
        assert future == future0
        callback0_called = True

    future0.add_done_callback(callback0)
    future0.set_result(1)
    assert callback0_called

    future1 = Future()
    callback1_called = False

    def callback1(future):
        nonlocal callback1_called
        assert future == future1  # pragma: no cover
        callback1_called = True  # pragma: no cover

    future1.add_done_callback(callback1)
    future1.remove_done_callback(callback1)
    future1.set_result(1)
    assert not callback1_called

    future2 = Future()

    def callback2(future):
        raise RuntimeError()

    future2.add_done_callback(callback2)

    with pytest.raises(RuntimeError):
        future2.set_result(1)

    future3 = Future()

    def callback3(future):
        raise RuntimeError()

    future3.add_done_callback(callback3)
    future3.add_done_callback(callback3)

    with pytest.raises(ExceptionGroup):
        future3.set_result(1)


async def test_add_done_callback_already_done():
    future = Future()
    future.set_result(0)
    callback_called = False

    def callback(future):
        nonlocal callback_called
        callback_called = True
        raise RuntimeError()

    with pytest.raises(RuntimeError):
        future.add_done_callback(callback)

    assert callback_called



from typing import Generic, TypeVar

from anyio import create_memory_object_stream


T = TypeVar("T")


class Queue(Generic[T]):
    def __init__(self, maxsize: int = 0):
        self._maxsize = maxsize
        max_buffer_size = float("inf") if maxsize <= 0 else maxsize
        self._send_stream, self._receive_stream = create_memory_object_stream[T](max_buffer_size=max_buffer_size)

    @property
    def maxsize(self) -> int:
        return self._maxsize

    def qsize(self) -> int:
        return self._send_stream.statistics().current_buffer_used

    def empty(self) -> bool:
        return self._send_stream.statistics().current_buffer_used == 0

    def full(self) -> bool:
        statistics = self._send_stream.statistics()
        return statistics.current_buffer_used == statistics.max_buffer_size

    async def put(self, item: T) -> None:
        await self._send_stream.send(item)

    def put_nowait(self, item: T) -> None:
        self._send_stream.send_nowait(item)

    async def get(self) -> T:
        return await self._receive_stream.receive()

    def get_nowait(self) -> T:
        return self._receive_stream.receive_nowait()

    def __del__(self) -> None:
        self._send_stream.close()
        self._receive_stream.close()



from __future__ import annotations

import inspect
from typing import Awaitable, Callable, Literal, TypeVar

import outcome

RetT = TypeVar("RetT")


def start_guest_run(
    async_fn: Callable[..., Awaitable[RetT]],
    *,
    run_sync_soon_threadsafe: Callable[[Callable[[], object]], object],
    done_callback: Callable[[outcome.Outcome[RetT]], object],
    run_sync_soon_not_threadsafe: Callable[[Callable[[], object]], object] | None = None,
    backend: Literal["asyncio"] | Literal["trio"] = "asyncio",
) -> None:
    if not inspect.iscoroutinefunction(async_fn):
        raise RuntimeError(f"Expected a coroutine function, got: {async_fn}")

    if backend == "asyncio":
        import aioguest

        aioguest.start_guest_run(
            async_fn(),
            run_sync_soon_threadsafe=run_sync_soon_threadsafe,
            done_callback=done_callback,
            run_sync_soon_not_threadsafe=run_sync_soon_not_threadsafe,
        )
    elif backend == "trio":
        import trio

        trio.lowlevel.start_guest_run(
            async_fn,
            run_sync_soon_threadsafe=run_sync_soon_threadsafe,
            done_callback=done_callback,
            run_sync_soon_not_threadsafe=run_sync_soon_not_threadsafe,
        )
    else:
        raise RuntimeError(f'Backend not supported: "{backend}"')



from __future__ import annotations

from sys import version_info
from typing import Callable, Generic, TypeVar

from anyio import Event, create_task_group
from anyio.abc import TaskGroup

from ._exceptions import CancelledError, InvalidStateError

if version_info < (3, 11):  # pragma: no cover
    from exceptiongroup import BaseExceptionGroup  # type: ignore[import-not-found]

T = TypeVar("T")


class Future(Generic[T]):
    _done_callbacks: list[Callable[[Future], None]]
    _exception: BaseException | None

    def __init__(self) -> None:
        self._result_event = Event()
        self._exception_event = Event()
        self._cancelled_event = Event()
        self._raise_cancelled_error = True
        self._done_callbacks = []
        self._done_event = Event()
        self._exception = None
        self._waiting = False

    def _call_callbacks(self) -> None:
        exceptions = []
        for callback in self._done_callbacks:
            try:
                callback(self)
            except BaseException as exc:
                exceptions.append(exc)
        if not exceptions:
            return
        if len(exceptions) == 1:
            raise exceptions[0]
        raise BaseExceptionGroup("Error while calling callbacks", exceptions)

    async def _wait_result(self, task_group: TaskGroup) -> None:
        await self._result_event.wait()
        task_group.cancel_scope.cancel()

    async def _wait_exception(self, task_group: TaskGroup) -> None:
        await self._exception_event.wait()
        task_group.cancel_scope.cancel()

    async def _wait_cancelled(self, task_group: TaskGroup) -> None:
        await self._cancelled_event.wait()
        task_group.cancel_scope.cancel()

    def cancel(self, raise_exception: bool = False) -> bool:
        if self._done_event.is_set() or self._cancelled_event.is_set():
            return False

        self._done_event.set()
        self._cancelled_event.set()
        self._raise_cancelled_error = raise_exception
        self._call_callbacks()
        return True

    def cancelled(self) -> bool:
        return self._cancelled_event.is_set()

    async def wait(self) -> T | None:
        if self._waiting:
            await self._done_event.wait()
        self._waiting = True
        if self._result_event.is_set():
            return self._result
        if self._exception_event.is_set():
            assert self._exception is not None
            raise self._exception
        if self._cancelled_event.is_set():
            if self._raise_cancelled_error:
                raise CancelledError

        async with create_task_group() as tg:
            tg.start_soon(self._wait_result, tg)
            tg.start_soon(self._wait_exception, tg)
            tg.start_soon(self._wait_cancelled, tg)

        if self._result_event.is_set():
            return self._result
        if self._exception_event.is_set():
            assert self._exception is not None
            raise self._exception
        if self._cancelled_event.is_set():
            if self._raise_cancelled_error:
                raise CancelledError

        return None  # pragma: nocover

    def done(self) -> bool:
        return self._done_event.is_set()

    def set_result(self, value: T) -> None:
        if self._done_event.is_set():
            raise InvalidStateError
        self._done_event.set()
        self._result = value
        self._result_event.set()
        self._call_callbacks()

    def result(self) -> T:
        if self._cancelled_event.is_set():
            raise CancelledError
        if self._result_event.is_set():
            return self._result
        if self._exception_event.is_set():
            assert self._exception is not None
            raise self._exception
        raise InvalidStateError

    def set_exception(self, value: BaseException) -> None:
        if self._done_event.is_set():
            raise InvalidStateError
        self._done_event.set()
        self._exception = value
        self._exception_event.set()
        self._call_callbacks()

    def exception(self) -> BaseException | None:
        if not self._done_event.is_set():
            raise InvalidStateError
        if self._cancelled_event.is_set():
            raise CancelledError
        return self._exception

    def add_done_callback(self, callback: Callable[[Future], None]) -> None:
        self._done_callbacks.append(callback)
        if self._done_event.is_set():
            callback(self)

    def remove_done_callback(self, callback: Callable[[Future], None]) -> int:
        count = self._done_callbacks.count(callback)
        for _ in range(count):
            self._done_callbacks.remove(callback)
        return count



import time

import pytest

from anyio import create_task_group, sleep
from anyioutils import Monitor


pytestmark = pytest.mark.anyio

async def test_busy():

    period = 0.01
    block = 0.05

    async def busy_task():
        while True:
            time.sleep(block)
            await sleep(0)

    async with Monitor(period) as monitor:
        async with create_task_group() as tg:
            tg.start_soon(busy_task)
            await sleep(period * 100)
            tg.cancel_scope.cancel()

    assert  (block / period) < monitor.result < (block / period) * 10


async def test_not_busy():

    period = 0.01

    async with Monitor(period) as monitor:
        await sleep(period * 100)

    assert  1 < monitor.result < 3
